; Definitions
(intent_def
  (intent) @name.definition.intent) @definition.intent

(slot_def
  (slot) @name.definition.slot) @definition.slot

(alias_def
  (alias) @name.definition.alias) @definition.alias

; References
(slot_ref
  (slot) @name.reference.slot) @reference.slot

(alias_ref
  (alias) @name.reference.alias) @reference.alias
